from enum import Enum
import base64
import json
from functools import partial
from typing import Union
from urllib.parse import urlparse
from aiocache import SimpleMemoryCache, cached
from tldextract import tldextract
from redis import asyncio
from spiders.realtime.rt_logger import logger, TraceID


class Operator(Enum):
    """
    操作符
    """
    equal = "equal"
    gt = "gt"
    lt = "lt"
    contains = "contains"
    i = "in"
    endswith = "endswith"
    startswith = "startswith"
    split = "split"
    index = "index"
    lower = "lower"  # 新增
    len = "len"
    code = "code"


file_suffix = [".3gp", ".aac", ".amr", ".app", ".avi", ".cab",
               ".doc", ".docx", ".exe", ".gbc", ".hme", ".jad",
               ".jar", ".m4a", ".mbm", ".mid", ".mjp", ".mmf", ".mp3",
               ".mp4", ".mpeg", ".mpkg", ".mrp", ".mtf", ".nes",
               ".nth", ".pdf", ".prc", ".pxl", ".rar",
               ".rm", ".sdt", ".sis", ".sisx", ".swf", ".thm", ".tsk", ".utz",
               ".wav", ".wma", ".wmv", ".zip", ".css", ".js",
               ".bmp", ".jpg", ".png", ".tif", ".gif", ".pcx", ".tga",
               ".exif", ".fpx", ".svg", ".psd", ".cdr", ".pcd", ".dxf",
               ".ufo", ".eps", ".ai", ".raw", ".wmf", ".webp",
               ".mov", ".asf", ".mkv", ".flv", ".f4v", ".xlsx", ".xls",
               ".iso", ".7z", ".gz", ".tgz", ".bz2", ".xz", ".z",
               ".exe", ".msi", ".apk", ".ipa", ".tmp", ".mdf",
               ".txt", ".rtf", ".ppt", ".pptx", ".bin"]

redirect_keyword = ["signin", "login", "register", "redirect", "registration", "authorize", "logout", "signup",
                    "sign-up", "sign-in"]
global_conf = {
    "*": [
        {
            "rule": {
                "path":
                    [[Operator.split.value, "/"], [Operator.index.value, -1],
                     [Operator.endswith.value, file_suffix + [suffix.upper() for suffix in file_suffix]]
                     ]
            },
            "extra": {"reason": ["manual"], "owner": ["421966"], "time": ["2024-12-20"], "md5": "xxxxxxxxx"}
        },
        {
            "rule": {
                "path":
                    [[Operator.split.value, "/"], [Operator.index.value, -1], [Operator.split.value, "."],
                     [Operator.index.value, 0], [Operator.lower.value],
                     [Operator.i.value, redirect_keyword + [keyword.upper() for keyword in redirect_keyword]]]

            },
            "extra": {"reason": ["manual"], "owner": ["421966"], "time": ["2024-12-20"], "md5": "yyyyyyyyyy"}
        },
        {
            "rule": {
                "url":
                    [[Operator.len.value], [Operator.gt.value, 2048]]

            },
            "extra": {"reason": ["manual"], "owner": ["421966"], "time": ["2024-03-12"]}
        }
    ]
}


class ReturnType(Enum):
    def __init__(self, code, msg):
        self.code = code
        self.msg = msg

    normal = (0, "normal")  # 正常 url
    black_host = (1, "hit black host")  # black host
    black_domain = (2, "hit black domain")  # black_domain
    white_host = (0, "hit white host")  # white_host
    white_domain = (0, "hit white domain")  # white_domain
    super_white_domain = (0, "hit super white domain")  # super_white_domain
    black_prefix = (3, "hit black prefix")  # black_prefix
    black_rule = (4, "hit black rule")  # black_rule

    url_parse_error = (100, "url parse error")  # URL 解析异常（解析失败认为 URL 不合法应该删除 #2024-12-25）
    filter_error = (0, "filter error")  # link-filter 处理异常（处理异常暂时不删除 #2024-12-25）


class Result:
    def __init__(self, hit: bool, msg: str, code: int, rule: Union[str, dict], extra: dict):
        """
        响应结果类型。hit 决定是否删除，code 提示规则类型。前期需要保持一致，也就是 hit=False 时，code 一定等于 0。
        后期 hit 和 code 可以不一致，比如 hit=False 时，code 可以不等于 0，表明虽然暂时不删除，但是需要指明因为哪个规则不删除。

        :param hit: 决定是否删除。True：删除、 False 不删除
        :param msg: 命中类型提示信息
        :param code: 命中类型
        :param rule: 命中规则
        :param extra: 返回额外信息：{"reason": "规则上线原因", "owner": "规则所属人", "time": "规则上线时间"}
        """
        self.hit = hit
        self.msg = msg
        self.type = code
        self.rule = rule
        self.extra = extra
        self.trace_id = TraceID.get_trace_id()

    def __str__(self):
        return json.dumps(self.__dict__)

    __repr__ = __str__

    to_dict = lambda self: self.__dict__


class LinkFilter:
    # 依赖规则表 1. odps 2. redis
    resource_link_filter_black_rule_name = "resource_link_filter_black_rule"
    resource_link_filter_black_prefix_name = "resource_link_filter_black_prefix"
    resource_link_filter_super_white_domain_name = "resource_link_filter_super_white_domain"
    resource_link_filter_black_host_name = "resource_link_filter_black_host"
    resource_link_filter_white_host_name = "resource_link_filter_white_host"
    resource_link_filter_white_domain_name = "resource_link_filter_white_domain"
    resource_link_filter_black_domain_name = "resource_link_filter_black_domain"

    @cached(ttl=86400, cache=SimpleMemoryCache)
    async def get(self, table_name, key):
        result = await self.client.hget(table_name, key)
        if result:
            return json.loads(result)
        return dict()

    def __init__(self, rds_host=None, rds_port=None, rds_password=None):
        self.client = asyncio.Redis(
            host=rds_host,
            port=rds_port,
            password=rds_password,
            db=4,
            max_connections=300
        )
        self.super_white_domain = partial(self.get, self.resource_link_filter_super_white_domain_name)  # 超级白名单
        self.white_domain = partial(self.get, self.resource_link_filter_white_domain_name)  # 域名白名单
        self.black_domain = partial(self.get, self.resource_link_filter_black_domain_name)  # 域名黑名单
        self.white_host = partial(self.get, self.resource_link_filter_white_host_name)  # host 白名单
        self.black_host = partial(self.get, self.resource_link_filter_black_host_name)  # host 黑名单
        self.black_rule = partial(self.get, self.resource_link_filter_black_rule_name)  # 规则黑名单 domain:rule
        self.black_prefix = partial(self.get, self.resource_link_filter_black_prefix_name)  # 前缀黑名单 host:prefix

    async def manual_evaluate(self, url, rules, serialize=True) -> Union[Result, dict]:
        """
        @param url: 链接
        @param rules: 规则
        {
            "chaxunmao.com": [
                {
                    "extra": {

                    },
                    "rule": {
                        "path": [
                            [
                                "split",
                                "/"
                            ],
                            [
                                "index",
                                -1
                            ],
                            [
                                "split",
                                "-"
                            ],
                            [
                                "len"
                            ],
                            [
                                "gt",
                                5
                            ]
                        ]
                    }
                }
            ]
        }
        @param serialize: 是否返回 json 序列化字符串
        """
        result = None
        try:
            domain, host, path, params = self.parse_url(url)
        except Exception as exc:
            # link-filter 内部逻辑错误
            result = Result(hit=False, msg=str(exc), code=ReturnType.filter_error.code, rule="", extra={})
        else:
            is_hit, rule, extra = await self.hit_black_rule(url, domain, host, path, params, rules)
            if is_hit:
                result = Result(hit=True, msg=ReturnType.black_rule.msg, code=ReturnType.black_rule.code,
                                rule=rule, extra=extra)
            else:
                # 所有规则均未命中并且没有报错
                result = Result(hit=False, msg=ReturnType.normal.msg, code=ReturnType.normal.code, rule="",
                                extra={})
        if result and serialize:
            return result.to_dict()
        return result

    async def evaluate(self, url):
        """
        是否命中规则？命中了哪一个规则？规则责任人是谁？规则在什么时间配置的？
        :param url: 链接
        :return: 返回 json 结果
        """
        try:
            domain, host, path, params = self.parse_url(url)
        except Exception as exc:
            # URL 解析异常
            logger.exception(f"url parse error", exc)
            return Result(hit=True, msg=str(exc), code=ReturnType.url_parse_error.code, rule=url, extra={}).to_dict()
        else:
            try:
                # # 判断是否在超级白名单中
                # extra = await self._in_super_white_domain(domain)
                # if extra:
                #     return Result(hit=False, msg=ReturnType.super_white_domain.msg,
                #                   code=ReturnType.super_white_domain.code, rule=domain, extra=extra).to_dict()

                # 判断是否命中人工规则
                is_hit, rule, extra = await self.hit_black_rule(url, domain, host, path, params)
                if is_hit:
                    return Result(hit=True, msg=ReturnType.black_rule.msg, code=ReturnType.black_rule.code,
                                  rule=rule, extra=extra).to_dict()

                # # 判断是否命中 prefix 黑名单
                # is_hit, rule, extra = await self.hit_black_prefix(url, host)
                # if is_hit:
                #     return Result(hit=True, msg=ReturnType.black_prefix.msg, code=ReturnType.black_prefix.code,
                #                   rule=rule, extra=extra).to_dict()
                #
                # # 判断是否命中 host 白名单
                # extra = await self._in_white_host(host)
                # if extra:
                #     return Result(hit=False, msg=ReturnType.white_host.msg, code=ReturnType.white_host.code,
                #                   rule=host, extra=extra).to_dict()
                #
                # # 判断是否命中 host 黑名单
                # extra = await self._in_black_host(host)
                # if extra:
                #     return Result(hit=True, msg=ReturnType.black_host.msg, code=ReturnType.black_host.code,
                #                   rule=host, extra=extra).to_dict()
                #
                # # 判断是否命中 domain 白名单
                # extra = await self._in_white_domain(domain)
                # if extra:
                #     return Result(hit=False, msg=ReturnType.white_domain.msg, code=ReturnType.white_domain.code,
                #                   rule=domain, extra=extra).to_dict()
                #
                # # 判断是否命中 domain 黑名单
                # extra = await self._in_black_domain(domain)
                # if extra:
                #     return Result(hit=True, msg=ReturnType.black_domain.msg, code=ReturnType.black_domain.code,
                #                   rule=domain, extra=extra).to_dict()
            except Exception as exc:
                # link-filter 内部逻辑错误
                logger.exception(f"link-filter internal error", exc)
                return Result(hit=False, msg=str(exc), code=ReturnType.filter_error.code, rule="", extra={}).to_dict()

            # 所有规则均未命中并且没有报错
            return Result(hit=False, msg=ReturnType.normal.msg, code=ReturnType.normal.code, rule="",
                          extra={}).to_dict()

    async def hit_black_prefix(self, url, host):
        prefixes = await self.black_prefix(host)
        host_with_http = f"http://{host}/"
        host_with_https = f"https://{host}/"
        if prefixes:
            for prefix_rules in prefixes:

                prefix = prefix_rules.get("rule")
                extra = prefix_rules.get("extra")

                if url.startswith(host_with_http + prefix) or url.startswith(host_with_https + prefix):
                    return True, f"{prefix}", extra
        return False, "", ""

    async def hit_black_rule(self, url, domain, host, path, params, rules=None):
        # 获取 domain 规则或者通配规则
        if rules:
            domain_rules = rules.get(domain) or rules.get("*")
        else:
            domain_rules: list = await self.black_rule(domain) or await self.black_rule("*")
            if not domain_rules:
                domain_rules = global_conf.get("*")
            else:
                domain_rules.extend(global_conf.get("*"))
        if domain_rules:
            for domain_rule in domain_rules:
                result = []
                rule = domain_rule.get("rule")
                extra = domain_rule.get("extra")

                url_rule = rule.get("url")
                host_rule = rule.get("host")
                path_rule = rule.get("path")
                params_rule = rule.get("params")

                rules = [("url", url, url_rule),
                         ("host", host, host_rule),
                         ("path", path, path_rule),
                         ("params", params, params_rule)
                         ]
                for name, part, part_rule in rules:
                    match = self.exec_rule(part, part_rule)
                    result.append(match)
                    if not match:
                        break
                if len(result) == len(rules) and all(result):
                    return True, rule, extra
        else:
            return False, "", {"msg": "no domain rules"}

        # 没有 domain 对应规则匹配失败，并返回没有匹配的原因
        return False, "", ""

    async def _in_super_white_domain(self, domain):
        return await self.super_white_domain(domain)

    async def _in_black_host(self, host):
        return await self.black_host(host)

    async def _in_white_host(self, host):
        return await self.white_host(host)

    async def _in_black_domain(self, domain):
        return await self.black_domain(domain)

    async def _in_white_domain(self, domain):
        return await self.white_domain(domain)

    @staticmethod
    def parse_url(url):
        parsed_url = urlparse(url)
        domain = tldextract.extract(url).registered_domain
        host = parsed_url.hostname
        path = parsed_url.path.strip("/")
        params = parsed_url.query
        return domain, host, path, params

    @staticmethod
    def exec_rule(part, rule: list):
        if not rule:
            return True
        result = part
        for r in rule:
            if r[0] == Operator.equal.value:
                result = result == r[1]
            elif r[0] == Operator.gt.value:
                result = result > r[1]
            elif r[0] == Operator.lt.value:
                result = result < r[1]
            elif r[0] == Operator.contains.value:
                result = r[1] in result
            elif r[0] == Operator.i.value:
                result = result in r[1]
            elif r[0] == Operator.endswith.value:
                if isinstance(r[1], list):
                    r[1] = tuple(r[1])
                result = result.endswith(r[1])
            elif r[0] == Operator.startswith.value:
                result = result.startswith(r[1])
            elif r[0] == Operator.split.value:
                result = result.split(r[1])
            elif r[0] == Operator.index.value:
                index = int(r[1])
                if index + 1 > len(result) and index != -1:
                    return False
                result = result[index]
            elif r[0] == Operator.len.value:
                result = len(result)
            elif r[0] == Operator.code.value:
                context = {}
                exec(base64.b64decode(r[1]).decode(), context)
                result = context["dynamic_function"](result)
            elif r[0] == Operator.lower.value:
                result = result.lower()
            else:
                raise RuntimeError(f"Unknown operator {r[0]}")
        return result



import os
import sys
from uuid import uuid4
import atexit
import platform

from multipledispatch import dispatch
from loguru._logger import Logger, Core
from contextvars import ContextVar

_trace_id: ContextVar[str] = ContextVar('x_trace_id', default="")  # 任务ID


class TraceID:
    """全链路追踪ID"""

    @staticmethod
    def set_trace_id(trace_id: str) -> ContextVar[str]:
        """设置全链路追踪ID
        Returns:
            ContextVar[str]: trace_id
        """
        _trace_id.set(trace_id)
        return _trace_id

    @staticmethod
    def get_trace_id() -> str:
        """获取全链路追踪ID
        Returns:
            str: trace_id
        """
        trace_id = _trace_id.get()
        if not trace_id:
            trace_id = str(uuid4()) + "-0000"
        return trace_id


def _logger_filter(record):
    record['trace_msg'] = f"[trace_id:{TraceID.get_trace_id()}]"
    return record['trace_msg']


loguru_logger = Logger(core=Core(),
                       exception=None,
                       depth=0,
                       record=False,
                       lazy=False,
                       colors=False,
                       raw=False,
                       capture=True,
                       patchers=[],
                       extra={}
                       )
atexit.register(loguru_logger.remove)

os_name = platform.system()

if os_name == "Darwin":
    colorize = True
else:
    colorize = False

log_level = os.getenv("RT_LOG_LEVEL") or "INFO"
if log_level not in ["OFF", "NONE"]:
    loguru_logger.add(sys.stdout,
                      format="<green><b>{time:YYYY-MM-DD HH:mm:ss.SSS}</b></green> <red>|</red> {level:<8} <red>|</red> "
                             "<yellow>{trace_msg}</yellow><blue><red>:</red>{name}<red>:</red>{function}<red>:</red>{line}"
                             "</blue> <red>-</red> {message}",
                      colorize=colorize,
                      filter=_logger_filter,
                      enqueue=True,
                      backtrace=True,
                      level=log_level
                      )
    loguru_logger.info(f"{os_name} realtime logger init success")
else:
    loguru_logger.remove()


class RealtimeLogger:
    PREFIX = "-*-JUHE-*- "

    @classmethod
    def info(cls, __message: str):
        loguru_logger.info(cls.PREFIX + str(__message))

    @dispatch(str, Exception)
    def error(self, __message: str, error: Exception):
        loguru_logger.exception(self.PREFIX + str(__message), error)

    @dispatch(str)
    def error(self, __message: str):
        loguru_logger.error(self.PREFIX + str(__message))

    @dispatch(Exception)
    def error(self, error: Exception):
        loguru_logger.error(self.PREFIX + str(error))

    @dispatch(str, Exception)
    def exception(self, __message: str, error: Exception):
        loguru_logger.exception(self.PREFIX + str(__message), error)

    @dispatch(str)
    def exception(self, __message: str):
        loguru_logger.error(self.PREFIX + str(__message))

    @dispatch(Exception)
    def exception(self, error: Exception):
        loguru_logger.exception(self.PREFIX, error)

    def log(self, __level, __message):
        loguru_logger.log(__level, self.PREFIX + str(__message))


logger = RealtimeLogger()



import setuptools

setuptools.setup(
    name="link-quality",
    version="0.0.6",
    author="daoya",
    license="MIT",
    python_requires=">=3.6",
    description="link-quality",
    install_requires=[
        "tldextract",
        "urllib3",
        "aioredis",
        "aiocache",
        "loguru",
        "multipledispatch"
    ],
    py_modules=[
        "spiders.realtime.link_filter",
        "spiders.realtime.normalize",
        "spiders.realtime.rt_logger",
    ],
    classifiers=["Programming Language :: Python :: 3"],
)



import json
import sys
import re
from enum import Enum

from collections import OrderedDict
from html import unescape
from typing import Union
from urllib.parse import urlparse, parse_qsl, urlunparse, ParseResult, quote_plus, unquote, unquote_plus

from aiocache import cached, SimpleMemoryCache
from redis import asyncio
from tldextract import tldextract

from spiders.realtime.link_filter import LinkFilter, Result
from spiders.realtime.rt_logger import logger


def urlencode(query, safe='', encoding=None, errors=None,
              quote_via=quote_plus):
    """
    重写urlencode函数 修复空参数末尾加=问题
    :param query:
    :param safe:
    :param encoding:
    :param errors:
    :param quote_via:
    :return:
    """
    if hasattr(query, "items"):
        query = query.items()
    else:
        try:
            if len(query) and not isinstance(query[0], tuple):
                raise TypeError
        except TypeError:
            ty, va, tb = sys.exc_info()
            raise TypeError("not a valid non-string sequence "
                            "or mapping object").with_traceback(tb)

    l = []
    for k, v in query:
        if isinstance(k, bytes):
            k = quote_via(k, safe)
        else:
            k = quote_via(str(k), safe, encoding, errors)

        if isinstance(v, bytes):
            v = quote_via(v, safe)
        else:
            v = quote_via(str(v), safe, encoding, errors)
        if not v:
            l.append(k)
        else:
            l.append(k + '=' + v)
    return '&'.join(l)


chinese_characters_re = re.compile(r'[\u4e00-\u9fff]+')


def params_unquote(encoded_str):
    """
    参数解码，去除 编码产物 amp;
    :param encoded_str:
    :return:
    """
    decoded_str = unquote(encoded_str)
    decoded_url = decoded_str
    while True:
        decoded_url_res = unescape(decoded_url)
        if decoded_url == decoded_url_res:
            break
        else:
            decoded_url = decoded_url_res

    # decoded_str = decoded_str.replace("amp;", "")
    return decoded_url


def keep_chinese(source_url, decoded_url):
    """
    维持原始Url的中文编码状态
    :param decoded_url:
    :return:
    """
    contains_chinese = chinese_characters_re.search(source_url) is not None
    if contains_chinese:
        # url 解码 ，维持中文
        decoded_url = unquote_plus(decoded_url)
    return decoded_url


def validate_url(url):
    """
    url 合法性验证
    :param url:
    :return:
    """
    if not url:
        return False

    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


class ReturnType(Enum):
    def __init__(self, code, msg):
        self.code = code
        self.msg = msg

    ## 规则匹配成功
    normal = (0, "normal")

    no_rules = (0, "no corresponding rules")  # 没有域名对应规则

    ## 异常
    url_parse_error = (1, "url parse error")  # URL 解析异常
    rule_not_match = (2, "rule not match")  # 规则引擎匹配失败
    normalize_error = (3, "normalize error")  # 归一化处理异常
    url_is_invalid = (4, "url is invalid")  # URL 非法


class NormalizeResult(Result):
    def __init__(self, hit: bool, msg: str, code: int, rule: Union[str, dict], extra: dict, url, normalize_url):
        super().__init__(hit=hit, msg=msg, code=code, rule=rule, extra=extra)
        self.url = url
        self.normalize_url = normalize_url


class Normalizer:
    def __init__(self, rds_host=None, rds_port=None, rds_password=None):
        self.client = asyncio.Redis(
            host=rds_host,
            port=rds_port,
            password=rds_password,
            db=4,
            max_connections=300,
        )
        self.link_filter = LinkFilter()
        self.normalize_table_key = "normalize:rules"

    async def manual_evaluate(self, url, rules=None, data=None):
        """
        归一化手动传入规则。如果有链接匹配规则先判断是否匹配，匹配后再判断再做归一化。如果 mirror 和 scheme 存在则替换为相应的值。
        @param url: 链接
        @param rules: 链接匹配规则
        @param data 归一化规则
        """

        keys = data.get("keys")
        mirror = data.get("mirror")
        new_scheme = data.get("scheme")
        try:

            scheme, netloc, path, params, query, domain = self.parse_url(url)
            query = params_unquote(query)
        except Exception as exc:
            # URL 解析异常
            logger.exception(f"url parse error", exc)
            return Result(hit=True, msg=ReturnType.url_parse_error.msg, code=ReturnType.url_parse_error.code,
                          rule=rules,
                          extra={"ori_url": url, "new_url": url, "data": data}
                          ).to_dict()
        else:
            # 判断是否命中前置规则（规则引擎）
            link_filter_result: Result = await self.link_filter.manual_evaluate(url, rules, serialize=False)

            rest_parts = []
            if link_filter_result.hit:
                parts = parse_qsl(query)
                for part, value in parts:
                    if part in keys:
                        if not int(keys[part]) == 1:
                            rest_parts.append((part, value))
                    else:
                        rest_parts.append((part, value))
            else:
                return Result(hit=False, msg=ReturnType.rule_not_match.msg, code=ReturnType.rule_not_match.code,
                              rule=rules,
                              extra={"ori_url": url, "new_url": url, "data": data}
                              ).to_dict()
            if mirror:
                netloc = mirror
            if new_scheme:
                scheme = new_scheme
            parse_result = ParseResult(scheme=scheme, netloc=netloc, path=path, params=params,
                                       query=urlencode(rest_parts), fragment=None)
            new_url = urlunparse(parse_result)
            return Result(hit=url != new_url, msg=ReturnType.normal.msg, code=ReturnType.normal.code, rule=rules,
                          extra={"ori_url": url, "new_url": new_url, "data": data}).to_dict()

    async def evaluate(self, url):
        """
        归一化接口。
        @param url: 链接
        """
        new_url, rule, extra, return_type = await self.check(url)
        return NormalizeResult(hit=url != new_url, msg=return_type.msg, code=return_type.code, rule=rule,
                               extra=extra, url=url, normalize_url=new_url).to_dict()

    @staticmethod
    def assemble_url(scheme: str, netloc: str, path: str, params: Union[str, None], query: dict[str, str]):
        return urlunparse(
            ParseResult(scheme=scheme, netloc=netloc, path=path, params=params, query=urlencode(query), fragment=None)
        )

    @cached(ttl=86400, cache=SimpleMemoryCache)
    async def get(self, domain):
        return await self.client.hget(self.normalize_table_key, domain)

    @staticmethod
    def parse_url(url):
        url_obj = urlparse(url)

        scheme = url_obj.scheme
        netloc = url_obj.netloc
        path = url_obj.path
        params = url_obj.params
        query = url_obj.query

        domain = tldextract.extract(url).registered_domain
        return scheme, netloc, path, params, query, domain

    @classmethod
    def normalize_offline(cls, scheme, netloc, path, params, query_dict):
        verified_url = cls.assemble_url(scheme=scheme, netloc=netloc, path=path, params=params, query=query_dict)
        logger.info(f"verified_url: {verified_url}")

    async def check(self, url: str):
        try:
            scheme, netloc, path, params, query, domain = self.parse_url(url)
        except Exception as exc:
            # URL 解析异常
            logger.exception(f"url parse error", exc)
            return url, dict(), dict(), ReturnType.url_parse_error

        if not all([scheme, netloc]):
            return url, dict(), dict(), ReturnType.url_is_invalid

        try:
            query = params_unquote(query)

            query_params = parse_qsl(query, keep_blank_values=True)

            rules = await self.get(domain)
            query_dict = dict()
            exclude_set = set()

            if not rules:
                # 新站点，直接走离线验证
                for k, v in query_params:
                    query_dict[k] = v
                    self.normalize_offline(scheme=scheme, netloc=netloc, path=path, params=params,
                                           query_dict=query_dict)
                return url, dict(), dict(), ReturnType.no_rules
            else:
                rules = json.loads(rules)

                for rule in rules:
                    data = rule.get("data")
                    extra = rule.pop("extra")

                    keys = data.get("keys")
                    mirror = data.get("mirror")
                    new_scheme = data.get("scheme")
                    link_filter_result = await self.link_filter.manual_evaluate(url, {domain: [rule]}, serialize=False)
                    if link_filter_result.hit:
                        for k, v in query_params:
                            check_code = keys.get(k)
                            if not check_code:
                                # 没有验证过的参数，当做必要参数 todo 离线验证，使用弹外能够使用的 rds 或者其他
                                query_dict[k] = v
                                self.normalize_offline(scheme=scheme, netloc=netloc, path=path, params=params,
                                                       query_dict=query_dict)
                            elif check_code == 3:
                                query_dict[k] = v
                            else:
                                exclude_set.add(k)

                        if mirror:
                            netloc = mirror
                        # if new_scheme:
                        #     scheme = new_scheme

                        # 重组url
                        new_url = self.assemble_url(scheme=scheme, netloc=netloc.lower(), path=path, params=params,
                                                    query=OrderedDict(sorted(query_dict.items())))

                        # 维持原始编码状态
                        res_url = keep_chinese(url, new_url)
                        return res_url, rule, extra, ReturnType.normal
                return url, dict(), dict(), ReturnType.rule_not_match
        except Exception as exc:
            logger.exception("url normalize error", exc)
            return url, dict(), dict(), ReturnType.normalize_error